
#define NODE0 (t_my_ref)0
#define NODE1 (t_my_ref)1

#include "dynare2.h"
#include "dynare3.h"
#include "d_tab.hh"
#include "moments.h"
#include "dynare_params.h"
#include "../sylv/cc/SylvException.h"
#include "../tl/cc/tl_exception.h"
#include "../kord/approximation.h"
#include "../kord/kord_exception.h"
#include "../kord/global_check.h"

#ifndef DYNVERSION
#define DYNVERSION "unknown"
#endif

// references the instance of Dynare class when unique parse is
// running
Dynare* dyn;


/**************************************************************************************/
/*       Dynare class                                                                 */
/**************************************************************************************/

Dynare::Dynare(const char* modname)
	: model_file(modname), ys(NULL), exe(NULL), param(NULL),
	  y(NULL), ex(NULL), md(1)
{
	parse();
	ys->print();
}

// Here is the copy constructor. Since all the structure do not have
// sensible copy constructors, we do everything from the scratch
// (namely from the model_file). We copy md only, parse the model and
// set the steady state to the original dynare.ys
Dynare::Dynare(const Dynare& dynare)
	: model_file(dynare.model_file), ys(NULL), exe(NULL), param(NULL),
	  y(NULL), ex(NULL), md(dynare.md)
{
	parse();
	ys = new Vector(*(dynare.ys));
}

void Dynare::parse()
{
	// absolute synchronization, no concurrency here
	SYNCHRO syn(0, "");

	// set dyn for lexicographic parser
	dyn = this;

	extern FILE *yyin;

	add_constant(INUMBER,"0");
	add_constant(INUMBER,"1");

	if (NULL==(yyin=fopen(model_file, "r"))) {
		fprintf(stderr, "Couldn't open \"%s\" file.\n", model_file);
		exit(1);
	}
	yyparse();

	sort_deriv_variables();

	derivator();

	dnl = new DynareNameList(*this);
	dsnl = new DynareStateNameList(*this, *dnl, nstat(), nboth(), npred(), nexog());
	denl = new DynareExogNameList(*this);

	y = new TwoDMatrix(-sm.maxlag+sm.maxlead+1,sm.endo_nbr);
	y->zeros();
	ex = new TwoDMatrix(-sm.maxlag+sm.maxlead+1,sm.exo_nbr);
	ex->zeros();

	if (sm.maxlag < -1) {
		fprintf(stderr, "Error: lags less than -1 not allowed.\n");
		exit(1);
	}
	if (sm.maxlead > 1) {
		fprintf(stderr, "Error: leads greater than 1 not allowed.\n");
		exit(1);
	}
}

Dynare::~Dynare()
{
	delete dnl;
	delete dsnl;
	delete denl;
	if (ys)
		delete ys;
	if (exe)
		delete exe;
	if (param)
		delete param;
	if (y)
		delete y;
	if (ex)
		delete ex;
}

void Dynare::writeMat4(FILE* fd, const char* prefix) const
{
	char tmp[100];
	sprintf(tmp, "%s_vars", prefix);
	getAllEndoNames().writeMat4(fd, tmp);
	getAllEndoNames().writeMat4Indices(fd, prefix);
	sprintf(tmp, "%s_state_vars", prefix);
	getStateNames().writeMat4(fd, tmp);
	sprintf(tmp, "%s_shocks", prefix);
	getExogNames().writeMat4(fd, tmp);
	getExogNames().writeMat4Indices(fd, prefix);
	sprintf(tmp, "%s_vcov_exo", prefix);
	getVcov().writeMat4(fd, tmp);
	TwoDMatrix aux(1,1);
	sprintf(tmp, "%s_nstat", prefix);
	aux.get(0,0) = nstat();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_npred", prefix);
	aux.get(0,0) = npred();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_nboth", prefix);
	aux.get(0,0) = nboth();
	aux.writeMat4(fd, tmp);
	sprintf(tmp, "%s_nforw", prefix);
	aux.get(0,0) = nforw();
	aux.writeMat4(fd, tmp);
}

void Dynare::solveDeterministicSteady(Vector& steady)
{
	Vector r1(1, *ex);
	r1 = (const Vector&) *exe;

	std::cout << "nlsolve\n";
	nlsolve(steady);
	std::cout << "Steady state\n";
	steady.print();
}

// evaluate system at given y_t=y_{t+1}=y_{t-1}, and given shocks x_t
void Dynare::evaluateSystem(Vector& out, const Vector& yy, const Vector& xx)
{
	TwoDMatrix yy_c(-sm.maxlag+sm.maxlead+1,sm.endo_nbr);
	TwoDMatrix xx_c(-sm.maxlag+sm.maxlead+1,sm.exo_nbr);
	for (int i = 0; i < -sm.maxlag+sm.maxlead+1; i++) {
		for (int j = 0; j < sm.endo_nbr; j++)
			yy_c.get(i,j) = yy[j];
		for (int j = 0; j < sm.exo_nbr; j++)
			xx_c.get(i,j) = xx[j];
	}
	vec_double g_v;
	vec_int g_c, g_r;
	Vector out_c(out.length());
	TensorContainer<FSSparseTensor> dummy(1);
	execute_tree(dummy, out_c, g_v, g_c, g_r, model.end(), 1-sm.maxlag, 0,
				 yy_c, xx_c);
	for (int j = 0; j < out.length(); j++)
		out[j] = out_c[j];
}

// evaluate system at given y^*_{t-1}, y_t, y^{**}_{t+1} and at
// exogenous x_t, all three vectors yym, yy, and yyp have the
// respective lengths of y^*_{t-1}, y_t, y^{**}_{t+1}
void Dynare::evaluateSystem(Vector& out, const Vector& yym, const Vector& yy,
							const Vector& yyp, const Vector& xx)
{
	TwoDMatrix yy_c(-sm.maxlag+sm.maxlead+1,sm.endo_nbr);
	TwoDMatrix xx_c(-sm.maxlag+sm.maxlead+1,sm.exo_nbr);
	for (int j = 0; j < yym.length(); j++)
		yy_c.get(0, nstat() + j) = yym[j];
	for (int j = 0; j < yy.length(); j++)
		yy_c.get(1, j) = yy[j];
	for (int j = 0; j < yyp.length(); j++)
		yy_c.get(2, nstat()+npred() + j) = yyp[j];
	for (int i = 0; i < -sm.maxlag+sm.maxlead+1; i++) {
		for (int j = 0; j < sm.exo_nbr; j++)
			xx_c.get(i,j) = xx[j];
	}
	vec_double g_v;
	vec_int g_c, g_r;
	Vector out_c(out.length());
	TensorContainer<FSSparseTensor> dummy(1);
	execute_tree(dummy, out_c, g_v, g_c, g_r, model.end(), 1-sm.maxlag, 0, yy_c, xx_c);
	for (int j = 0; j < out.length(); j++)
		out[j] = out_c[j];
}

void Dynare::calcDerivatives(const Vector& yy, const Vector& xx)
{
	// set y and ex
	for (int i = 0; i < -sm.maxlag+sm.maxlead+1; i++) {
		Vector ri(i, *y);
		ri = yy;
	}
	for (int i = 0; i < -sm.maxlag+sm.maxlead+1; i++) {
		Vector ri(i, *ex);
		ri = xx;
	}

	Vector zz(sm.endo_nbr);
	vec_double g_v;
	vec_int g_c, g_r;
	md.clear();
	execute_tree(md, zz, g_v, g_c, g_r, model.end(), 1-sm.maxlag, 1);
}

void Dynare::calcDerivativesAtSteady()
{
	Vector xx(nexog());
	xx.zeros();
	calcDerivatives(*ys, xx);
}





bool operator == (const struct s_node & s1, const struct s_node &s2){
  //  std::cout << s1.arg1 << " " << s1.arg2 << " " << s1.arg2 << " " << s2.arg2 << " " << s1.code << " " << s2.code << "\n";
  return (s1.arg1 == s2.arg1 && s1.arg2==s2.arg2 && s1.code == s2.code);
}

bool operator !=(const struct s_node & s1, const struct s_node &s2){
  return (s1.arg1 != s2.arg1 || s1.arg2!=s2.arg2 || s1.code != s2.code);
}

bool variable_comp::operator()(const struct s_variable_key & s1, const struct s_variable_key & s2)
{
	if (s1.lag < s2.lag)
		return 1;
	else if (s1.lag > s2.lag)
		return 0;
	else
		return strcmp(s1.name,s2.name) < 0;
}

t_my_ref Dynare::add_node_check(int code, t_my_ref arg1, t_my_ref arg2)
  /* Code contains the code of the operator
     arg1 contains the address of the first passed argument
     arg2 contains the address of the second passed argument
     if argi=0 then result=NODE0
     if argi=1 then result=NODE1
  */
{
#ifdef __MINGW32__
	if (code == ACOSH || code == ASINH || code == ATANH) {
		fprintf(stderr, "cosh, sinh, and tanh not implemented for MinGW\n");
		exit(1);
	}
#endif

  switch(code)
    {
    case PLUS:
      if (arg1 != NODE0 && arg2 != NODE0)
	{
	  return add_node(PLUS,arg1,arg2);
	}
      else if (arg1 != NODE0)
	{
	  return arg1;
	}
      else if (arg2 != NODE0)
	{
	  return arg2;
	}
      else
	{
	  return NODE0;
	}
      break;
    case MINUS:
      if (arg1 != NODE0 && arg2 != NODE0)
	{
	  return add_node(MINUS,arg1,arg2);
	}
      else if (arg1 != NODE0)
	{
	  return arg1;
	}
      else if (arg2 != NODE0)
	{
	  return add_node(MINUS,0,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TIMES:
      if (arg1 != NODE0 && arg1 != NODE1 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(TIMES,arg1,arg2);
	}
      else if (arg1 != NODE0 && arg1 != NODE1 && arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg2 != NODE0 && arg2 != NODE1 && arg1 == NODE1)
	{
	  return arg2;
	}
      else if (arg2 == NODE1 && arg1 == NODE1)
	{
	  return NODE1;
	}
      else
	{
	  return NODE0;
	}
      break;
    case DIVIDE:
      if (arg1 != NODE0 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(DIVIDE,arg1,arg2);
	}
      else if (arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg1 == NODE0 && arg2 != NODE0)
	{
	  return NODE0;
	}
      else
	{
	  std::cerr << "DIVIDE 0/0 non available\n";
	  exit(-1);
	}
      break;
    case EXP:
      if (arg2 != NODE0)
	{
	  return add_node(EXP,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case LOG:
      if (arg2 != NODE0)
	{
	  return add_node(LOG,arg1,arg2);
	}
      else
	{
	  std::cerr << "log(0) isn't available\n";
	  exit(-1);
	}
      break;
    case LOG10:
      if (arg2 != NODE0)
	{
	  return add_node(LOG10,arg1,arg2);
	}
      else
	{
	  std::cerr << "log10(0) isn't available\n";
	  exit(-1);
	}
      break;
    case COS:
      if (arg2 != NODE0)
	{
	  return add_node(COS,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case SIN:
      if (arg2 != NODE0)
	{
	  return add_node(SIN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TAN:
      if (arg2 != NODE0)
	{
	  return add_node(TAN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ACOS:
	  return add_node(ACOS,arg1,arg2);
      break;
    case ASIN:
      if (arg2 != NODE0)
	{
	  return add_node(ASIN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ATAN:
      if (arg2 != NODE0)
	{
	  return add_node(ATAN,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case COSH:
      if (arg2 != NODE0)
	{
	  return add_node(COSH,arg1,arg2);
	}
      else
	{
	  return NODE1;
	}
      break;
    case SINH:
      if (arg2 != NODE0)
	{
	  return add_node(SINH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case TANH:
      if (arg2 != NODE0)
	{
	  return add_node(TANH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ACOSH:
	  return add_node(ACOSH,arg1,arg2);
      break;
    case ASINH:
      if (arg2 != NODE0)
	{
	  return add_node(ASINH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case ATANH:
      if (arg2 != NODE0)
	{
	  return add_node(ATANH,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case SQRT:
      if (arg2 != NODE0)
	{
	  return add_node(SQRT,arg1,arg2);
	}
      else
	{
	  return NODE0;
	}
      break;
    case POWER:
      if (arg1 != NODE0 && arg2 != NODE0 && arg2 != NODE1)
	{
	  return add_node(POWER,arg1,arg2);
	}
      else if (arg2 == NODE1)
	{
	  return arg1;
	}
      else if (arg2 == NODE0)
	{
	  return NODE1;
	}
      else
	{
	  return NODE0;
	}
      break;
    default:
      std::cerr << "missing code in add_node_check " << code << "\n";
      exit(-1);
    }
}

t_my_ref Dynare::add_node(int code, t_my_ref arg1, t_my_ref arg2)
{
  struct s_node node;

  /* for commutative operators, argument addresses are ordered */
    if ((code == PLUS || code == TIMES) && arg1 > arg2)
    {
      t_my_ref temp=arg2;
      arg2=arg1;
      arg1=temp;
    }

  node.code = code;
  node.arg1 = arg1;
  node.arg2 = arg2;
  if (arg1 != NODE0 && arg1 != NODE1)
	  ++(arg1->ref);
  if (arg2 != NODE0 && arg2 != NODE1)
	  ++(arg2->ref);

  tm_it address = find(model.begin(),model.end(),node);

  if (address == model.end() )
    {
      node.ref = 0;

      model.push_back(node);
      //      return (--model.end())._M_node;
      return &model.back();
    }
  else
    {
      return &(*address);
    }
}

void Dynare::add_result(int arg1, t_my_ref arg2)
{
  if (arg2 != NODE0 && arg2 != NODE1) ++(arg2->ref);

  struct s_node node;

  node.code = EQUAL_SIGN;
  node.arg1 = NODE0;
  node.arg2 = arg2;
  node.ref = arg1;
  node.no = 0;
  model.push_back(node);
  if (arg1 == sm.endo_nbr-1)
    {
      end_resid = model.end();
      end_resid--;
    }
}

void Dynare::add_deriv(int arg1, t_my_ref arg2, int arg3, int arg4, int arg5,
					   const IntSequence & intseq)
{
  if (arg2 != NODE0 && arg2 != NODE1) ++(arg2->ref);

  struct s_node node;

  node.code = EQUAL_SIGN;
  node.arg1 = NODE0;
  node.arg2 = arg2;
  node.ref = arg1;

  static int index;
  node.no = ++index; 
  s_deriv deriv(arg4, arg3, arg5, intseq);
  deriv_col.insert(t_h_deriv::value_type(index, deriv));

  model.push_back(node);
}

void Dynare::add_symbol(const char *name, int type)
{
  symbols[name].type=type;
  if (type == 0)
    {
      symbols[name].nbr=sm.exo_nbr;
      sm.exo_nbr++;
    }
  else if (type == 1)
    {
      symbols[name].nbr=sm.endo_nbr;
      sm.endo_nbr++;
    }
  else if (type == 3)
    {
      symbols[name].nbr=sm.param_nbr;
      sm.param_nbr++;
    }
}

void Dynare::add_symbol_range(char *name1, char *name2, int type)
{
  int i1, i2, i3;
  char name[200],*p1,*p2;

  p1=name1;
  p2=name2;
  i1=0;
  while (*p1 == *p2)
    {
      p1++;
      p2++;
      i1++;
    }

  if (*p1 >= '0' & *p1 <= '9')
    {
      if (sscanf(p1,"%d",&i2) != 1)
	{
	  fprintf(stderr,"VAR, VAREXO or VARDEF: error in name range expression\n");
	  return;
	}    
      if (sscanf(p2,"%d",&i3) != 1)
	{
	  fprintf(stderr,"VAR, VAREXO or VARDEF: error in name range expression\n");
	  return;
	}    
      if (i3 < i2)
	{
	  std::cerr << "VAR, VAREXO or VARDEF: error in name range expression\n";
	  return;
	}
    
      strncpy(name,name1,i1);
      while (i2 <= i3)
	{
	  sprintf(name+i1,"%d",i2);
	  add_symbol(name, type);
	  i2++;
	}
    }
  else
    {
      std::cerr << "VAR, VAREXO or VARDEF: error in name range expression\n";
      return;
    }
  return;
}

t_my_ref Dynare::add_constant(int code, const char *s)
{
  if (strcmp(s,"0")==0) return NODE0;
  if (strcmp(s,"1")==0) return NODE1;

  s_variable_key v;
  v.name = s;
  v.lag = 0;
  type_variable_table::iterator address = variables.find(v);

  if (address == variables.end() )
    {
      add_symbol(s,3);
      set_constant(s);
      return add_variable(s,0,0);
    }
  return variables[v].address;
}


t_my_ref Dynare::add_variable(const char *s,int lag, int dummy)
{
  type_symbol_table::iterator i = symbols.find(s);
  if (i == symbols.end())
    {
      std::cerr << "Unknown symbol in model: " << s;
      exit(1);
    }

  if (i->second.type == 0 && lag != 0)
    {
      std::cerr << "Exogenous shocks must be in current period. Use another endogenous variable.\n";
	}
  s_variable_key v;
  v.name = i->first;
  v.lag = lag;
  if (lag < sm.maxlag)
    {
      sm.maxlag = lag;
    }
  else if (lag > sm.maxlead)
    {
      sm.maxlead = lag;
    }

  type_variable_table::iterator address = variables.find(v);

  if (address == variables.end() ) {
      variables[v].type = i->second.type;
      if (i->second.type == 1) {
		  std::cout << s << " " << lag << "\n";
		  sm.dyn_endo_nbr++;
	  }
      variables[v].nbr = i->second.nbr;
      if (dummy == 0) {
		  s_node node;
		  node.code = 0;
		  node.arg1 = NODE0;
		  node.arg2 = NODE0;
		  model.push_back(node);
		  //	  variables[v].address = (--model.end())._M_node;
		  variables[v].address = &(model.back());
	  } else {
		  variables[v].address = (t_my_ref)0;
	  }
  }

  if (lag != 0) {
	  add_variable(s, 0, dummy);
  }

  return variables[v].address;
}

void Dynare::derivator()
{
  std::cout << "derivatives\n";

  // initialize derivatives of variables 
  int nbr_deriv_var = 0;
  for (t_var_deriv_index::iterator i=var_deriv_index.begin(); i != var_deriv_index.end(); i++)
    {
      if ( ((*i)->second).address != (t_my_ref)0 )
	{
	  //	  ((tm_it)(((*i)->second).address))->derivatives[nbr_deriv_var] = NODE1;
	  ((((*i)->second).address))->derivatives[nbr_deriv_var] = NODE1;
	}
      ++nbr_deriv_var;
    }

  tm_it k2 = model.begin();
  --k2;
  vec_int deriv_block(nbr_deriv_var);
  deriv_block[0] = 0;
  for (int i=1; i < nbr_deriv_var; i++)
    {
      deriv_block[i] = 1;
    }
    
  // loop on order of derivation
  for(int i=0; i < deriv_order; i++)
    {
      std::cout << "  order " << i+1 << "\n";
      // updates nbr of derivatives to skip
      tm_it k1 = k2;
      ++k1;
      k2=model.end();
      --k2;
      // loop on variables of derivation
      int offset = 0;
      for (int j=0; j < nbr_deriv_var; j++)
	{
	  std::cout << "    var " << j << "\n";
	  if (i > 0)
	    {
	      deriv_block[j]=deriv_block[j]*(nbr_deriv_var+i-j-1)/i;
	      offset += deriv_block[j];
	    }
	  // loop on nodes
	  for (tm_it k=k1;; k++)
	    {
	      if (k->code == 0) continue; // this is a leaf

	      t_my_ref a, d1, d2, t1, t2, t3, t4, t5;
	      a = k->arg1;
	      if (a == NODE0 || a == NODE1) d1 = NODE0;
	      else
		{
		  //		  t_derivatives dd = ((tm_it)a)->derivatives;
		  t_derivatives& dd = a->derivatives;
		  t_derivatives::iterator id = dd.find(j);
		  if (id == dd.end()) d1 = NODE0;
		  else d1 = dd[j];
		}
	      a = k->arg2;
	      if (a == NODE0 || a == NODE1) d2 = NODE0;
	      else
		{
		  //		  t_derivatives dd = ((tm_it)(k->arg2))->derivatives;
		  t_derivatives& dd = (k->arg2)->derivatives;
		  t_derivatives::iterator id = dd.find(j);
		  if (id == dd.end()) d2 = NODE0;
		  else d2 = dd[j];
		} 
	      switch (k->code)
		{
		case PLUS:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(PLUS,d1,d2);
		    }
		  break;
		case MINUS:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(MINUS,d1,d2);
		    }
		  break;
		case TIMES:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,d1,k->arg2);
		      t2 = add_node_check(TIMES,d2,k->arg1);
		      k->derivatives[j]=add_node_check(PLUS,t1,t2);
		    }
		  break;
		case DIVIDE:
		  if (d1 != NODE0 || d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,d1,k->arg2);
		      t2 = add_node_check(TIMES,d2,k->arg1);
		      t3 = add_node_check(MINUS,t1,t2);
		      t4 = add_node_check(TIMES,k->arg2,k->arg2);
		      k->derivatives[j]=add_node_check(DIVIDE,t3,t4);
		    }
		  break;
		case EXP:
		  if (d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(TIMES,d2,&(*k));
		    }
		  break;

		case LOG:
		  if (d2 != NODE0)
		    {
		      k->derivatives[j]=add_node_check(DIVIDE,d2,k->arg2);
		    }
		  break;
		case LOG10:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(EXP,NODE1,NODE1);
		      t2 = add_node_check(LOG10,NODE1,t1);
		      t3 = add_node_check(DIVIDE,d2,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,t2,t3);
		    }
		  break;
		case COS:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(SIN,NODE0,k->arg2);
		      t2 = add_node_check(MINUS,NODE0,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case SIN:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(COS,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case TAN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,&(*k),&(*k));
		      t2 = add_node_check(PLUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case ACOS:
		  if (d2 != NODE0)
		    {
		  t1 = add_node_check(SIN,NODE0,&(*k));
		  t2 = add_node_check(DIVIDE,d2,t1);
		  k->derivatives[j]=add_node_check(MINUS,NODE0,t2);
		    }
		  break;
		case ASIN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(COS,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ATAN:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,k->arg2,k->arg2);
		      t2 = add_node_check(PLUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t2);
		    }
		  break;
		case COSH:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(SINH,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case SINH:
		  if (d2 != NODE0)
		    {
		      t2 = add_node_check(COSH,k->arg1,k->arg2);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case TANH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,&(*k),&(*k));
		      t2 = add_node_check(MINUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(TIMES,d2,t2);
		    }
		  break;
		case ACOSH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(SINH,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ASINH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(COSH,NODE0,&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case ATANH:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(TIMES,k->arg2,k->arg2);
		      t2 = add_node_check(MINUS,NODE1,t1);
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t2);
		    }
		  break;
		case SQRT:
		  if (d2 != NODE0)
		    {
		      t1 = add_node_check(PLUS,&(*k),&(*k));
		      k->derivatives[j]=add_node_check(DIVIDE,d2,t1);
		    }
		  break;
		case POWER:
		  if (d1 != NODE0)
		    {
		      if (d2==0)
			{
			  t1 = add_node_check(MINUS,k->arg2,NODE1);
			  t2 = add_node_check(POWER,k->arg1,t1);
			  t3 = add_node_check(TIMES,k->arg2,t2);
			  k->derivatives[j]=add_node_check(TIMES,d1,t3);
			}
		      else
			{
			  t1 = add_node_check(LOG,NODE0,k->arg1);
			  t2 = add_node_check(TIMES,d2,t1);
			  t3 = add_node_check(TIMES,d1,k->arg2);
			  t4 = add_node_check(DIVIDE,t3,k->arg1);
			  t5 = add_node_check(PLUS,t2,t4);
			  k->derivatives[j]=add_node_check(TIMES,t5,&(*k));
			}
		    }
		  break;
		case EQUAL_SIGN:
		  // skip symmetric derivatives
		  if (i == 0 && d2 != NODE0)
		    {
		      IntSequence intseq(1);
		      intseq[0] = j;
		      add_deriv(k->ref,d2,j+1,j,i,intseq);
		    }
		  else {
			  t_h_deriv::const_iterator it = deriv_col.find(k->no);
			  if (it != deriv_col.end() && (*it).second.var >= j && d2 != NODE0)
		    {
		      IntSequence intseq(i+1);
		      for(int is=0;is<i;++is) intseq[is] = (*it).second.intseq[is];
		      intseq[i] = j;
		      add_deriv(k->ref,d2,offset+(*it).second.col,j,i,intseq);
		    }
		  }
		  break;
		default:
		  std::cout << "Missing code in DERIV:" << k->code;
		}                    /* end switch */
	      if (k == k2) break;
	    }                        /* end for k=k1..k2 */
	  if (i == 0 && j == sm.dyn_endo_nbr - 1)
	    {
	      end_endo_deriv = model.end();
	      --end_endo_deriv;
	    }
	}                            /* end for j=0...nbr_deriv_var */

      for(tm_rit k = model.rbegin(); k != (tm_rit)k2; k++)
	{
	  if (k->code != 0 && k->code != EQUAL_SIGN && k->ref == 0){
	    //	    if (k->arg1 != NODE0 && k->arg1 != NODE1) --(((tm_it)k->arg1)->ref);
	    //	    if (k->arg2 != NODE0 && k->arg2 != NODE1) --(((tm_it)k->arg2)->ref);
	    if (k->arg1 != NODE0 && k->arg1 != NODE1) --((k->arg1)->ref);
	    if (k->arg2 != NODE0 && k->arg2 != NODE1) --((k->arg2)->ref);
	  }
	}
      tm_it k3 = model.end();
      for(tm_it k = k2; k != k3;)
	{
	  tm_it k1 = k;
	  ++k1;
	  if (k->code != 0 && k->code != EQUAL_SIGN && k->ref == 0){
       	    model.erase(k);
	  }
	  k=k1;
	}
      if ( i == 0)
	{
	  end_first_deriv = model.end();
	  --end_first_deriv;
	}
    }                                /* end for i=0...deriv_order */
}

void Dynare::init_var()
{
	if (ys)
		delete ys;
	ys = new Vector(sm.endo_nbr);
	ys->zeros();
}
  
void Dynare::init_varexo()
{
	if (exe)
		delete exe;
	exe = new Vector(sm.exo_nbr);
	exe->zeros();
}
  
void Dynare::init_param()
{
	if (param)
		delete param;
	param = new vec_double(sm.param_nbr);
}

void Dynare::set_deriv_order(const char *number)
{
  std::istringstream is(number);
  is >> deriv_order;
  printf("deriv_order: %s=%d\n", number, deriv_order);
} 

double Dynare::get_param(const char *name)
{
  type_symbol_table::iterator i = symbols.find(name);
  if (i == symbols.end())
    {
      std::cout << "Unknown parameter: " << name;
      exit(1);
    }
  if ((i->second).type != 3)
    {
      std::cout << "Non-parameter used in VCOV expression: " << name;
      exit(1);
    }

  return (*param)[(i->second).nbr];
}

DynareNameList::DynareNameList(const Dynare& dynare)
{
	for (t_var_deriv_index::const_iterator i = dynare.var_deriv_index.begin();
		 i != dynare.var_deriv_index.end(); i++) {
		if (((*i)->first).lag == 0 && ((*i)->second).type == 1)
			names.push_back(((*i)->first).name);
    }
}

DynareStateNameList::DynareStateNameList(const Dynare& dynare, const DynareNameList& dnl,
										 int nstat, int npred,
										 int nboth, int nexog)
{
	// add predetermined variables
	for (int i = nstat; i < nstat+npred+nboth; i++)
		names.push_back(dnl.getName(i));

	// add exogenous variables
	for (t_var_deriv_index::const_iterator i = dynare.var_deriv_index.begin();
		 i != dynare.var_deriv_index.end(); i++) {
		if (((*i)->first).lag == 0 && ((*i)->second).type == 0)
			names.push_back(((*i)->first).name);
	}
}

DynareExogNameList::DynareExogNameList(const Dynare& dynare)
{
	for (t_var_deriv_index::const_iterator i = dynare.var_deriv_index.begin();
		 i != dynare.var_deriv_index.end(); i++) {
		if (((*i)->first).lag == 0 && ((*i)->second).type == 0)
			names.push_back(((*i)->first).name);
    }
}

int main(int argc, char** argv)
{
	DynareParams params(argc, argv);
	if (params.help) {
		params.printHelp();
		return 0;
	}
	if (params.version) {
		printf("Dynare++ v. %s. Copyright (C) 2004,2005 Ondra Kamenik, Michel Juillard.\n",
			   DYNVERSION);
		printf("Dynare++ comes with ABSOLUTELY NO WARRANTY and is distributed under\n");
		printf("General Public License, see http://www.gnu.org/license/gpl.html\n");
		return 0;
	}
	THREAD_GROUP::max_parallel_threads = params.num_threads;

	Dynare dynare(params.modname);

	try {
		tls.init(dynare.order(),
				 dynare.nstat()+2*dynare.npred()+3*dynare.nboth()+
				 2*dynare.nforw()+dynare.nexog());

		// make journal name
		char jname[500];
		sprintf(jname, "%s.jnl", params.basename);
		Journal journal(jname);

		// write information on variables to journal
		{
			JournalRecord jr(journal);
			jr << "The model has " << dynare.nstat() << " static, " << dynare.nboth()+dynare.npred()
			   << " predetermined, " << dynare.nboth()+dynare.nforw() << " frwd-looking, including "
			   << dynare.nboth() << " at both times" << endrec;
		}

		Approximation app(dynare, journal, params.num_steps);
		app.walkStochSteady();


		// open mat file
		char matfile[100];
		sprintf(matfile, "%s.mat", params.basename);
		FILE* matfd;
		if (NULL == (matfd=fopen(matfile, "wb"))) {
			fprintf(stderr, "Couldn't open %s for writing.\n", matfile);
			exit(1);
		}

		dynare.writeMat4(matfd, params.prefix);

		char ss_matrix_name[100];
		sprintf(ss_matrix_name, "%s_steady_states", params.prefix);
		ConstTwoDMatrix(app.getSS()).writeMat4(matfd, ss_matrix_name);

		// check the approximation
		if (params.check_along_path || params.check_along_shocks
			|| params.check_on_ellipse) {
			GlobalChecker gcheck(app, THREAD_GROUP::max_parallel_threads, journal);
			if (params.check_along_shocks)
				gcheck.checkAlongShocksAndSave(matfd, params.prefix,
											   params.getCheckShockPoints(),
											   params.getCheckShockScale(),
											   params.check_evals);
			if (params.check_on_ellipse)
				gcheck.checkOnEllipseAndSave(matfd, params.prefix,
											 params.getCheckEllipsePoints(),
											 params.getCheckEllipseScale(),
											 params.check_evals);
			if (params.check_along_path)
				gcheck.checkAlongSimulationAndSave(matfd, params.prefix,
												   params.getCheckPathPoints(),
												   params.check_evals);
		}

		// construct the decision rule and simulate
		const DecisionRule<KOrder::fold>& dr = app.getDecisionRule();
		dr.writeMat4(matfd, params.prefix);
		RandomGenerator::initSeed(params.seed);
		SimResultsStats res(dynare.numeq(), params.num_per);
		RandomShockRealization sr(dynare.getVcov());
		res.simulate(params.num_sim, dr, dynare.getSteady(), sr, journal);
		res.writeMat4(matfd, params.prefix);

		// impulse response functions
		IRFResults irf(dynare, dr, res, journal);
		irf.writeMat4(matfd, params.prefix);

		fclose(matfd);

	} catch (const KordException& e) {
		printf("Caugth Kord exception: ");
		e.print();
	} catch (const TLException& e) {
		printf("Caugth TL exception: ");
		e.print();
	} catch (SylvException& e) {
		printf("Caught Sylv exception: ");
		e.printMessage();
	}

  return 0;
}
