// $Id: dynare3.h 431 2005-08-16 15:41:01Z kamenik $
// Copyright 2005, Ondra Kamenik

#ifndef DYNARE3_H
#define DYNARE3_H

#include "../tl/cc/t_container.h"
#include "../tl/cc/sparse_tensor.h"
#include "../kord/decision_rule.h"
#include "../kord/dynamic_model.h"

#include "dynare2.h"
#include "moments.h"

#include <vector>

class Dynare;

class DynareNameList : public NameList {
	vector<const char*> names;
public:
	DynareNameList(const Dynare& dynare);
	int getNum() const
		{return (int)names.size();}
	const char* getName(int i) const
		{return names[i];}
};

class DynareStateNameList : public NameList {
	vector<const char*> names;
public:
	DynareStateNameList(const Dynare& dynare, const DynareNameList& dnl,
						int nstat, int npred, int nboth, int nexog);
	int getNum() const
		{return (int)names.size();}
	const char* getName(int i) const
		{return names[i];}
};

class DynareExogNameList : public NameList {
	vector<const char*> names;
public:
	DynareExogNameList(const Dynare& dynare);
	int getNum() const
		{return (int)names.size();}
	const char* getName(int i) const
		{return names[i];}
};

// The following only encapsulates every variable which used to be
// declared as global. It provides some methods to use the functions
// of non-linear solver, parser, and evaluator safely and repeatedly.

class Dynare : public DynamicModel {
	friend class DynareNameList;
	friend class DynareStateNameList;
	friend class DynareExogNameList;
	friend class Moments;
	const char* model_file;
	type_model model;
	type_symbol_table symbols;
	type_variable_table variables;
	t_var_deriv_index var_deriv_index;
	t_h_deriv deriv_col;
	std::vector<s_mixed> mixed;
	model_struct sm;
	Vector* ys;
	Vector* exe;
	vec_double* param;
	TwoDMatrix* y;
	TwoDMatrix* ex;
	tm_it end_resid;
	tm_it end_endo_deriv;
	tm_it end_first_deriv;
	int deriv_order;
	TensorContainer<FSSparseTensor> md;
	Moments moments;
	DynareNameList* dnl;
	DynareStateNameList* dsnl;
	DynareExogNameList* denl;
public:
	Dynare(const char* modname);
	Dynare(const Dynare& dyn);
	DynamicModel* clone() const
		{return new Dynare(*this);}
	~Dynare();
	int nstat() const
		{return sm.static_nbr;}
	int nboth() const
		{return sm.c_both_nbr;}
	int npred() const
		{return sm.s_lagged_nbr - sm.c_both_nbr;}
	int nforw() const
		{return sm.fwrd_nbr - sm.c_both_nbr;}
	int nexog() const
		{return sm.exo_nbr;}
	int order() const
		{return deriv_order;}

	const NameList& getAllEndoNames() const
		{return *dnl;}
	const NameList& getStateNames() const
		{return *dsnl;}
	const NameList& getExogNames() const
		{return *denl;}

	const TwoDMatrix& getVcov() const
		{return moments.getVcov();}
	const TensorContainer<FSSparseTensor>& getModelDerivatives() const
		{return md;}
	const Vector& getSteady() const
		{return *ys;}
	Vector& getSteady()
		{return *ys;}

	// these are used by lexical parses, otherwise dangerous
	Moments& getMoments()
		{return moments;}
	model_struct& getSM()
		{return sm;}

	// should not normally belong to public interface, but these are
	// used by lexical parser
	void init_var();
	void init_varexo();
	void init_param();
	void add_symbol(const char *name, int type);
	void add_symbol_range(char *name1, char *name2, int type);
	t_my_ref add_node(int code, t_my_ref arg1, t_my_ref arg2);
	void add_result(int arg1, t_my_ref arg2);
	t_my_ref add_constant(int code, const char *s);
	t_my_ref add_variable(const char *s,int lag, int dummy);
	void set_param(const char*, char*);
	void set_param(const char*, char*, int);
	void set_var(const char*, char*);
	void set_var(const char*, char*, int);
	void set_deriv_order(const char *number);
	double get_param(const char *name);

	// here is true public interface
	void solveDeterministicSteady(Vector& steady);
	void solveDeterministicSteady()
		{solveDeterministicSteady(*ys);}
	void evaluateSystem(Vector& out, const Vector& yy, const Vector& xx);
	void evaluateSystem(Vector& out, const Vector& yym, const Vector& yy,
						const Vector& yyp, const Vector& xx);
	void calcDerivatives(const Vector& yy, const Vector& xx);
	void calcDerivativesAtSteady();

	void writeMat4(FILE* fd, const char* prefix) const;
private:
	void parse();
	void sort_deriv_variables();
	void derivator();
	void nlsolve(Vector& steady);
	int lnsrch(Vector & x, double f, Vector & g, Vector & p,
			   double stpmax, Vector & fvec);
	// executes with y and ex from object
	void execute_tree(TensorContainer<FSSparseTensor>&, Vector&, vec_double &, vec_int &,
					  vec_int &, tm_it, int, int);
	// executes with y and ex provided
	void execute_tree(TensorContainer<FSSparseTensor>&, Vector &, vec_double &, vec_int &,
					  vec_int &, tm_it, int, int, const TwoDMatrix&  yy, const TwoDMatrix& xx);
	void state(t_state& state_vec);
	void jacobian(Vector& zz, TwoDMatrix& jacob);

	t_my_ref add_node_check(int code, t_my_ref arg1, t_my_ref arg2);
	void add_deriv(int arg1, t_my_ref arg2, int arg3, int arg4, int arg5,
				   const IntSequence & intseq);
	void set_constant(const char*);
};

#endif

// Local Variables:
// mode:C++
// End:
