// $Id: dynare_params.cc 431 2005-08-16 15:41:01Z kamenik $
//Copyright 2004, Ondra Kamenik

#include "dynare_params.h"

#include <getopt.h>
#include <stdio.h>
#include <string.h>

const char* help_str = 
"usage: dynare++ [--help] [options] <model file>\n"
"\n"
"    --help               print this message and return\n"
"    --version            print version and return\n"
"\n"
"options:\n"
"    --per <num>          number of periods simulated [100]\n"
"    --sim <num>          number of simulations [80]\n"
"    --steps <num>        steps towards stoch. SS [0=deter.]\n"
"    --prefix <string>    prefix of variables in Mat-4 file [\"dyn\"]\n"
"    --seed <num>         random number generator seed [934098]\n"
"    --threads <num>      number of max parallel threads [2]\n"
"    --check pesPES       check model residuals [no checks]\n"
"                         lower/upper case switches off/on\n"
"                           pP  checking along simulation path\n"
"                           eE  checking on ellipse\n"
"                           sS  checking along shocks\n"
"   --check-evals <num>   max number of evals per residual [1000]\n"
"   --check-num <num>     number of checked points [10]\n"
"   --check-scale <num>   scaling of checked points [2.0]\n"
"\n\n";

DynareParams::DynareParams(int argc, char** argv)
	: modname(NULL), num_per(100), num_sim(80), num_threads(2), num_steps(0),
	  prefix("dyn"), seed(934098), check_along_path(false), check_along_shocks(false),
	  check_on_ellipse(false), check_evals(1000), check_num(10), check_scale(2.0),
	  help(false), version(false)
{
	basename[0]='\0';

	if (argc == 1 || !strcmp(argv[1],"--help")) {
		help = true;
		return;
	}
	if (argc == 1 || !strcmp(argv[1],"--version")) {
		version = true;
		return;
	}

	modname = argv[argc-1];
	argc--;

	struct option const opts [] = {
		{"periods", required_argument, NULL, opt_per},
		{"per", required_argument, NULL, opt_per},
		{"simulations", required_argument, NULL, opt_sim},
		{"sim", required_argument, NULL, opt_sim},
		{"prefix", required_argument, NULL, opt_prefix},
		{"threads", required_argument, NULL, opt_threads},
		{"steps", required_argument, NULL, opt_steps},
		{"seed", required_argument, NULL, opt_seed},
		{"check", required_argument, NULL, opt_check},
		{"check-scale", required_argument, NULL, opt_check_scale},
		{"check-evals", required_argument, NULL, opt_check_evals},
		{"check-num", required_argument, NULL, opt_check_num},
		{"help", no_argument, NULL, opt_help},
		{"version", no_argument, NULL, opt_version},
		{NULL, 0, NULL, 0}
	};

	int ret;
	int index;
	while (-1 != (ret = getopt_long(argc, argv, "", opts, &index))) {
		switch (ret) {
		case opt_per:
			if (1 != sscanf(optarg, "%d", &num_per))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_sim:
			if (1 != sscanf(optarg, "%d", &num_sim))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_prefix:
			prefix = optarg;
			break;
		case opt_threads:
			if (1 != sscanf(optarg, "%d", &num_threads))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_steps:
			if (1 != sscanf(optarg, "%d", &num_steps))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_seed:
			if (1 != sscanf(optarg, "%d", &seed))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_check:
			processCheckFlags(optarg);
			break;
		case opt_check_scale:
			if (1 != sscanf(optarg, "%lf", &check_scale))
				fprintf(stderr, "Couldn't parse float %s, ignored\n", optarg);
			break;
		case opt_check_evals:
			if (1 != sscanf(optarg, "%d", &check_evals))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_check_num:
			if (1 != sscanf(optarg, "%d", &check_num))
				fprintf(stderr, "Couldn't parse integer %s, ignored\n", optarg);
			break;
		case opt_help:
			help = true;
			break;
		case opt_version:
			version = true;
			break;
		case '?':
			fprintf(stderr, "Unknown option, ignored\n");
			break;
		}
	}

	// make basename
	strcpy(basename, modname);
	int i = strlen(basename)-1;
	while (i >= 0 && basename[i] != '.') i--;
	if (i > 0)
		basename[i] = '\0';
}

void DynareParams::printHelp() const
{
	printf("%s", help_str);
}

void DynareParams::processCheckFlags(const char* flags)
{
	for (unsigned int i = 0; i < strlen(flags); i++) {
		switch (flags[i]) {
		case 'p':
			check_along_path = false;
			break;
		case 'P':
			check_along_path = true;
			break;
		case 'e':
			check_on_ellipse = false;
			break;
		case 'E':
			check_on_ellipse = true;
			break;
		case 's':
			check_along_shocks = false;
			break;
		case 'S':
			check_along_shocks = true;
			break;
		default:
			fprintf(stderr, "Unknow check type selection character <%c>, ignored.\n", flags[i]);
		}
	}
}

// Local Variables:
// mode:C++
// End:
